import logging

from typing import Union, List, Optional

from ldap import ALREADY_EXISTS, UNWILLING_TO_PERFORM, NO_SUCH_OBJECT
from ldap.ldapobject import SimpleLDAPObject as LdapObject

from . import _ldaps, exceptions
from ._object_classes import User, Group, Computer, create_records
from ._query import object_exists, get, modify, _build_query_params

logger = logging.getLogger(__name__)
class_plural = "groups"
object_class = "group"


@create_records(Group)
def _get(
    connection: LdapObject, search_base: Optional[str] = None, search_scope: Optional[int] = None, **query_params
) -> List[Group]:
    return get(
        f"(objectClass={object_class})", connection, search_base=search_base, search_scope=search_scope, **query_params
    )


def _create(connection: LdapObject, distinguished_name: str, name: str, group_type: str, **extra_params) -> None:
    """
    From: https://adsearch-winzero.blogspot.com/2006/04/object-group-attribute-grouptype.html

    The groupType attribute returns the type of group. However the returned value is in RAW format. AdSearch conversion
    tool will convert the groupType result, or for further reference see the possible return values below:

    -2147483646 ~ Global Security Group
    -2147483644 ~ Local Security Group
    -2147483643 ~ BuiltIn Group
    -2147483640 ~ Universal Security Group

    2 ~ Global Distribution Group
    4 ~ Local Distribution Group
    8 ~ Universal Distribution Group
    """
    try:
        group_type = {"global": "-2147483646", "domain_local": "-2147483644", "universal": "-2147483640"}[group_type]
    except KeyError:
        raise exceptions.ParameterError(f"Supplied group_type {group_type} is not supported.")

    attr_dict: dict = {
        "distinguishedName": bytes(distinguished_name, "utf-8"),
        "sAMAccountName": bytes(name, "utf-8"),
        "name": bytes(name, "utf-8"),
        "objectClass": [b"top", bytes(object_class, "utf-8")],
        "groupType": bytes(group_type, "utf-8"),
    }
    attr_dict.update(_ldaps._encode_attributes(extra_params))
    return _ldaps._create(connection, distinguished_name, attr_dict)


@object_exists(_get, class_plural)
def _delete(connection: LdapObject, distinguished_name: str) -> None:
    return _ldaps._delete(connection, distinguished_name)


@object_exists(_get, class_plural)
def _move(connection: LdapObject, distinguished_name: str, new_parent: str) -> None:
    return _ldaps._move(connection, distinguished_name, new_parent)


@object_exists(_get, class_plural)
def _members(
    connection: LdapObject, distinguished_name: str, nested: Optional[bool] = False
) -> List[Union[Group, User]]:
    def _resolve_group_member(group_distinguished_name):
        logger.info(f"resolving group member {group_distinguished_name}")
        return next(
            iter(_ldaps._search(connection, _build_query_params({"distinguishedName": group_distinguished_name})))
        )

    def _generate_member_record(raw_group_member):
        member_class = raw_group_member["objectClass"]
        if b"group" in member_class:
            return Group(**raw_group_member)
        elif b"computer" in member_class:
            # This *must* come before the user test, because computers are also users -_-
            return Computer(**raw_group_member)
        elif b"user" in member_class:
            return User(**raw_group_member)
        else:
            logger.error(f"cannot generate record from the following attributes: {raw_group_member}")
            raise exceptions.OperationFailed(
                f"Failed to create either a user or group record from object with objectClass {member_class}"
            )

    def _get_flat_members(group):
        """
        - Get all 'member' distinguishedNames from the group.
        - If the group has > 1500 members, the member_range key should be used instead.
        - In the above scenario, you can never see all the members.
        - Work out if each member is a user or a group.
        - Create the relevant namedtuple.
        """
        if all([not group.member, not group.member_range]):
            return []

        member_list = group.member if group.member else group.member_range

        # Force it to be a list just in case there's a single member
        return [
            _generate_member_record(_resolve_group_member(member_dn))
            for member_dn in ([member_list] if not isinstance(member_list, list) else member_list)
        ]

    def _get_nested_members(group):
        """
        - Use the special LDAP_MATCHING_RULE_IN_CHAIN (OID 1.2.840.113556.1.4.1941) to get nested group members.
        - Work out if each member is a user or a group.
        - Create the relevant namedtuple.
        """
        nested_members_query = (
            "(&(|(objectClass=user)(objectClass=person)(objectClass=group))(!(objectClass=computer))"
            f"(memberOf:1.2.840.113556.1.4.1941:={group.distinguishedName}))"
        )
        nested_members = _ldaps._search(connection, nested_members_query)
        return (
            []
            if not nested_members
            else (
                [
                    # Because we're using the low-level _ldaps._search function here, none of the attribute decoding
                    # will have happened yet. For that reason we have to collapse and cast [b'distinguishedName'] down
                    # to "distinguishedName" before passing it into the next layer.
                    _generate_member_record(_resolve_group_member(member["distinguishedName"][0].decode("utf-8")))
                    for member in nested_members
                ]
            )
        )

    group = next(iter(_get(connection, distinguishedName=distinguished_name)))
    return _get_nested_members(group) if nested else _get_flat_members(group)


@object_exists(_get, class_plural)
def _modify(
    connection: LdapObject, distinguished_name: str, key: str, value: Union[int, str, bytes], operation: str
) -> None:
    return modify(connection, distinguished_name, key, value, operation)


def _modify_members(
    connection: LdapObject, distinguished_name: str, member_distinguished_name: str, operation: str
) -> None:
    try:
        return _modify(connection, distinguished_name, "member", bytes(member_distinguished_name, "utf-8"), operation)
    except ALREADY_EXISTS:
        # This will only happen if the operation is 'add'.
        raise exceptions.ObjectExists(
            f"Supplied member '{member_distinguished_name}' is already in group '{distinguished_name}'"
        )
    except UNWILLING_TO_PERFORM:
        if operation == "add":
            raise exceptions.OperationFailed(
                "Adding a new member to the group failed. This may be because the type "
                "of the member and the group are incompatible."
            )
        raise exceptions.OperationFailed(
            f"Supplied member '{member_distinguished_name}' is not in group '{distinguished_name}'."
        )
    except NO_SUCH_OBJECT as e:
        # If NO_SUCH_OBJECT happens here, it's because the new member does not
        # exist. The call to _modify ensures the group exists via the _group_exists
        # decorator, but the supplied member is not checked.
        message = f"Supplied member could not be found using distinguishedName '{member_distinguished_name}'."
        try:
            message += f" Found the following closest matches: '{e.args[0].get('matched', 'N/A')}'."
        except IndexError:
            pass
        raise exceptions.NoSuchObject(message)
