import logging
import functools

from typing import Union

from . import exceptions
from ._attribute_codecs import AttributeCodecs, EncodedLdapBytes, EncodedQueryString, DecodedLdapString

logger = logging.getLogger(__name__)

ENCODED_ATTRIBUTES = {"logonHours", "objectGUID", "objectSid", "thumbnailPhoto"}  # noqa


class ProtectedString(str):
    r"""
    A type of `str` that, when passed into a search function as a value, will **not be escaped** when
    building the LDAP query.

    Use this class if you want to search with a special value that should not be escaped.

    Example - an already-encoded `objectGUID` in string format which would normally be escaped:

      .. code-block:: python

        with active_directory_client() as connection:
            user = users.search(
            connection,
            objectGUID=ProtectedString(r"\39\1b\5d\ef\30\33\8d\44\89\c8\41\ea\63\5f\ef\2a")
        )

    """


def _codec(func):
    @functools.wraps(func)
    def _fn(name, value):
        try:
            return func(name, value)
        # Any encoding/decoding exceptions will be raised by the decorated function, we don't need to catch them here.
        except AttributeError:
            # Name does not exist in attribute_codecs
            raise exceptions.ParameterError(f"No codec available for attribute {name}.")

    return _fn


@_codec
def query_encode_attribute(
    name: str, value: Union[str, bytes, DecodedLdapString, EncodedLdapBytes]
) -> EncodedQueryString:
    r"""
    Encode an LDAP attribute value into a format usable in an LDAP query.

    Example 1 - Encoding a known objectGUID string:

    .. code-block:: python

        object_guid_string = "ef5d1b39-3330-448d-89c8-41ea635fef2a"
        query_encode_attribute("objectGUID", object_guid_string)

    Example 2 - Encoding the objectGUID of a previously retrieved user record:

    .. code-block:: python

        query_encode_attribute("objectGUID", ad_user.objectGUID)

    Example 3 - Encoding raw LDAP bytes objectGUID value:

    .. code-block:: python

        query_encode_attribute("objectGUID", b'9\x1b]\xef03\x8dD\x89\xc8A\xeac_\xef*')

    Output for all examples:

    .. code-block:: bash

        EncodedQueryString(objectGUID='\39\1b\5d\ef\30\33\8d\44\89\c8\41\ea\63\5f\ef\2a')

    :param name: The name of the LDAP attribute. This value is case sensitive.
    :param value: The value to encode. Can be `bytes`, `str`, `DecodedLdapString`, or `EncodedLdapBytes`.

    :returns: The LDAP attribute value in query string format (`EncodedQueryString`).

    :raises exceptions.ParameterError: if the supplied `name` does not have an associated
        encoder, **or** the codec does not contain a query encoder, **or** the supplied `value` is invalid. The
        exception message will provide clarity.
    """
    return getattr(AttributeCodecs, name).query_encode(value)


@_codec
def bytes_encode_attribute(name: str, value: Union[str, DecodedLdapString]) -> EncodedLdapBytes:
    r"""
    Encode an LDAP attribute value into raw LDAP bytes.

    As LDAP queries do not support bytes values, you should rarely need to use this function from a PPA task.

    Example - Encoding the objectGUID of a previously retrieved user record:

    .. code-block:: python

        object_guid_string = "ef5d1b39-3330-448d-89c8-41ea635fef2a"
        bytes_encode_attribute("objectGUID", object_guid_string)

    Output:

    .. code-block:: bash

        EncodedLdapBytes(objectGUID=b'9\x1b]\xef03\x8dD\x89\xc8A\xeac_\xef*')


    :param name: The name of the LDAP attribute. This value is case sensitive.
    :param value: The value to encode. Can be `str` or `DecodedLdapString`.

    :returns: The LDAP attribute value in raw LDAP bytes format (`EncodedLdapBytes`).

    :raises exceptions.ParameterError: if the supplied `name` does not have an associated
        encoder, **or** the codec does not contain a bytes encoder, **or** the supplied `value` is
        invalid. The exception message will provide clarity.
    """
    return getattr(AttributeCodecs, name).bytes_encode(value)


@_codec
def decode_attribute(name: str, value: bytes) -> DecodedLdapString:
    r"""
    Decode an LDAP attribute value into a human readable string.

    As this function is called automatically when creating user and computer records, you will rarely need to use it
    from inside a PPA task.

    Example - Decoding an objectGUID from raw LDAP bytes:

    .. code-block:: python

        object_guid_bytes = b'9\x1b]\xef03\x8dD\x89\xc8A\xeac_\xef*'
        bytes_encode_attribute("objectGUID", object_guid_bytes)

    Output:

    .. code-block:: bash

        DecodedLdapString(objectGUID='ef5d1b39-3330-448d-89c8-41ea635fef2a')

    :param name: The name of the LDAP attribute. This value is case sensitive.
    :param value: The value to decode. Must be `bytes`.

    :returns: The LDAP attribute value in raw LDAP bytes format (`EncodedLdapBytes`).

    :raises exceptions.ParameterError: if the supplied `name` does not have an associated
        encoder, **or** the codec does not contain a decoder, **or** the supplied `value` is invalid. The
        exception message will provide clarity.
    """
    return getattr(AttributeCodecs, name).decode(value)
