"""
Users

This module contains actions related to Active Directory Users.
"""
import logging
import uuid
import datetime
import functools

from typing import List, Dict, Optional, Union, Any

from ppa import ui

from playbook_tools import tables, process
from active_directory import helpers, users, groups, exceptions, _translators, _ldaps, _users, exceptions as sdk_exceptions

from . import domain
from .exceptions import NoUserFound, InvalidAttribute
from .records import User, Group, DomainController, PasswordExpiry, UACPropertyFlags, AttributeSet
from ._wrappers import ldap_function, idempotent_create, idempotent_remove, process_user, process_group, bool_to_search_scope


USER_TABLE_HEADER = ["Common Name", "SAM Account Name", "Email Address", "Enabled", "Locked"]
USER_TABLE_FIELDS = ["cn", "sAMAccountName", "mail", "is_enabled", "is_locked"]


logger = logging.getLogger(__name__)


def _process_header_and_fields(header: Optional[List[str]], fields: Optional[List[str]]) -> tuple:
    if all((header is not None, fields is None)):
        ui.output_error("Custom header input was supplied without fields, default settings will be used.")
        header, fields = (None, None)
    elif all((header is None, fields is not None)):
        ui.output_error("Custom fields input was supplied without a header, default settings will be used.")
        header, fields = (None, None)
    if all((header is not None, fields is not None, len(header or []) != len(fields or []))):
        ui.output_error("Custom table header & fields do not match, default settings will be used.")
        header, fields = (None, None)
    return (header, fields)


@ldap_function
def search(
    domain_controller: DomainController,
    search_params: Dict[str, Any],
    search_base: Optional[str] = None,
    one_level: Optional[bool] = False,
) -> process.SearchResult:
    """Search for users using LDAP attributes & values.

!!! question "Minimum Plugin Version: 6.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __search_params:__ a dictionary containing [user keys](../inputs_outputs#user) & values to use in the search

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

    - __one_level:__ set to `true` to search one level (see [here](../advanced_queries/#search-scopes) for more information)

??? success "Output"

    A [SearchResult](../../../searches/#search-result-dictionary) containing:

    - **total**: the number of [Users](../inputs_outputs/#user) found

    - **all**: a list of [Users](../inputs_outputs/#user) found

    - **first**: the first [User](../inputs_outputs/#user) found

    - **last**: the last [User](../inputs_outputs/#user) found

??? warning "Search Result Format"

    The search result format was updated in version 6.0.0 of this plugin.

    It is not compatible with playbooks written against previous versions.

    See [here](/searches/#using-search-results) for more information.

???+ warning "Search Speed"

    This operation can be slow on a large domain.

    Consider targeting the search using `search_base` where possible.

??? example

    Searching for all users in the __Users__ CN whose `sAMAccountName` starts with __admin__:

    ```YAML linenums="1"
    - active_directory.users.search:
        search_params:
          sAMAccountName: admin*
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: users
    ```

!!! tip

    You can use `*` as a wildcard at the end of search values.
"""
    return process.search_result(
        list(
            map(
                process_user,
                users.search(
                    domain_controller, search_base=search_base, search_scope=bool_to_search_scope(one_level), **search_params
                ),
            )
        )
    )


@ldap_function
def members_of(
    domain_controller: DomainController, group_distinguishedNames: List[str], search_base: Optional[str] = None
) -> List[User]:
    """Find users who are members of __any__ of the supplied groups.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __group_distinguishedNames:__ a list of group distinguishedNames

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

??? success "Output"

    A list of [User Dictionaries](../inputs_outputs/#user).

???+ warning "Search Speed"

    This operation can be slow on a large domain, especially if multiple groups are supplied.

    Consider targeting the search using `search_base` where possible.

??? example "Single Group"

    Finding users in the __Remote Desktop Users__ group:

    ```YAML linenums="1" hl_lines="3"
    - active_directory.users.members_of:
        group_distinguishedNames:
          - CN=Remote Desktop Users,CN=Builtin,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: users
    ```

??? example "Targeted Search & Multiple Groups"

    Finding all users who are:

    - In the default __Users__ CN

    - In _either_ the __Account Operators__ _or_ __Remote Desktop Users__ groups

    ```YAML linenums="1" hl_lines="3 4 5"
    - active_directory.users.members_of:
        group_distinguishedNames:
          - CN=Remote Desktop Users,CN=Builtin,DC=Example,DC=Domain,DC=Com
          - CN=Account Operators,CN=Users,DC=Example,DC=Domain,DC=Com
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: users
    ```
"""
    return list(
        map(
            process_user,
            users.members_of(domain_controller, group_distinguished_names=group_distinguishedNames, search_base=search_base),
        )
    )


@ldap_function
def not_members_of(
    domain_controller: DomainController, group_distinguishedNames: List[str], search_base: Optional[str] = None
) -> List[User]:
    """Find users who are not members of __any__ of the supplied groups.

!!! question "Minimum Plugin Version: 3.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __group_distinguishedNames:__ a list of group distinguishedNames

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

??? success "Output"

    A list of [User Dictionaries](../inputs_outputs/#user).

???+ warning "Search Speed"

    This operation can be slow on a large domain, especially if multiple groups are supplied.

    Consider targeting the search using `search_base` where possible.

??? example "Single Group"

    Finding all users not in the __Remote Desktop Users__ group:

    ```YAML linenums="1" hl_lines="3"
    - active_directory.users.not_members_of:
        group_distinguishedNames:
          - CN=Remote Desktop Users,CN=Builtin,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: users
    ```

??? example "Targeted Search & Multiple Groups"

    Finding all users who are:

    - In the default __Users__ CN

    - Not in _either_ the __Account Operators__ _or_ __Remote Desktop Users__ groups

    ```YAML linenums="1" hl_lines="3 4 5"
    - active_directory.users.not_members_of:
        group_distinguishedNames:
          - CN=Remote Desktop Users,CN=Builtin,DC=Example,DC=Domain,DC=Com
          - CN=Account Operators,CN=Users,DC=Example,DC=Domain,DC=Com
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
      load:
        domain_controller: domain_controller_info
      save: users
    ```
"""
    return list(
        map(
            process_user,
            users.not_members_of(domain_controller, group_distinguished_names=group_distinguishedNames, search_base=search_base),
        )
    )


@ldap_function
def by_samaccountname(domain_controller: DomainController, sAMAccountName: str, search_base: Optional[str] = None) -> User:
    """Get a user by its 'sAMAccountName' attribute.

!!! question "Minimum Plugin Version: 7.10.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __sAMAccountName:__ The user's `sAMAccountName`

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

??? success "Output"

    A single [User Dictionary](../inputs_outputs/#user).

??? example
    ```YAML linenums="1"
    - active_directory.users.by_samaccountname:
        sAMAccountName: example.user
      load:
        domain_controller: domain_controller_info
      save: example_user
    ```
"""
    try:
        return process_user(users.search(domain_controller, search_base=search_base, sAMAccountName=sAMAccountName)[0])
    except IndexError:
        raise NoUserFound(f"No user found with sAMAccountName '{sAMAccountName}'.") from None


@ldap_function
def by_distinguishedname(domain_controller: DomainController, distinguishedName: str, search_base: Optional[str] = None) -> User:
    """Get a user by its 'distinguishedName' attribute.

!!! question "Minimum Plugin Version: 7.10.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ The user's `distinguishedName`

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

??? success "Output"
    A single [User Dictionary](../inputs_outputs/#user).

??? example
    ```YAML linenums="1"
    - active_directory.users.by_distinguishedname:
        distinguishedName: cn=Example User,CN=Users,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller_info
      save: example_user
    ```
"""
    try:
        return process_user(users.search(domain_controller, search_base=search_base, distinguishedName=distinguishedName)[0])
    except IndexError:
        raise NoUserFound(f"No user found with distinguishedName '{distinguishedName}'.") from None


@ldap_function
def get_free_samaccountname(domain_controller: DomainController, sAMAccountNames: List[str]) -> Optional[str]:
    """Takes a list of sAMAccountNames & finds the first that is currently available.

This can use useful in new joiner tasks when multiple naming formats are acceptable.

!!! question "Minimum Plugin Version: 7.9.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __sAMAccountNames:__ a list of sAMAccountNames to check

??? success "Output"

    Either the first available __sAMAccountName__ or __null__ if they are _all_ in use.

??? example
    ```YAML linenums="1"
    - active_directory.users.get_free_samaccountname:
        sAMAccountNames:
          - example.user
          - e.user
          - exampleuser
      load:
        domain_controller: domain_controller
      save: available_name
    ```
"""
    for name in sAMAccountNames:
        logger.info(f"checking whether sAMAccountName {name} is available")
        try:
            by_samaccountname(domain_controller, name)
            logger.info(f"sAMAccountName {name} is not available")
        except NoUserFound:
            logger.info(f"sAMAccountName {name} is available")
            return name
    logger.info("none of the supplied sAMAccountNames are available")
    return None


@ldap_function
def get_free_userprincipalname(domain_controller: DomainController, userPrincipalNames: List[str]) -> Optional[str]:
    """Takes a list of userPrincipalNames & finds the first that is currently available.

This can use useful in new joiner tasks when multiple naming formats are acceptable.

!!! question "Minimum Plugin Version: 7.9.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __userPrincipalNames:__ a list of userPrincipalNames to check

??? success "Output"

    Either the first available __userPrincipalName__ or __null__ if they are _all_ in use.

??? example
    ```YAML linenums="1"
    - active_directory.users.get_free_userprincipalname:
        userPrincipalNames:
          - example.user@example.domain
          - e.user@example.domain
          - exampleuser@example.domain
      load:
        domain_controller: domain_controller
      save: available_name
    ```
"""
    for name in userPrincipalNames:
        logger.info(f"checking whether userPrincipalName {name} is available")
        if search(domain_controller, search_params={"userPrincipalName": name})["total"] == 0:
            logger.info(f"userPrincipalName {name} is available")
            return name
        logger.info(f"userPrincipalName {name} is not available")
    logger.info("none of the supplied userPrincipalNames are available")
    return None


@ldap_function
def get_interactive(
    domain_controller: DomainController,
    title: Optional[str] = None,
    search_attribute: Optional[str] = "sAMAccountName",
    exclude: Optional[Dict[str, Union[str, int]]] = None,
    search_base: Optional[str] = None,
) -> User:
    """Start an interactive search for a user.

!!! question "Minimum Plugin Version: 2.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __title:__ a title displayed to the Task Operator (defaults to `Get User Account`)

    - __search_attribute:__ an attribute from the following list:

        - _sAMAccountName_ (default value)
        - _cn_
        - _mail_

    - __exclude:__ a dictionary containing LDAP attribute names & regular expressions to test their values

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

??? success "Output"

    A single [User Dictionary](../inputs_outputs/#user).

!!! tip "Automatic Wildcards"

    When using this action all provided search terms will have a wildcard appended.

??? example

    - Interactively searching for a user inside the builtin __Users__ CN

    - All users whose `sAMAccountName` contains __admin__ are excluded from the results

    ```YAML linenums="1"
    - active_directory.users.get_interactive:
        search_base: CN=Users,DC=Example,DC=Domain,DC=Com
        exclude:
          sAMAccountName: .*admin.*
      load:
        domain_controller: domain_controller_info
      save: user
    ```
    """
    logger.info("starting interactive user search")
    while True:
        try:
            return process_user(
                helpers.get_user(
                    domain_controller,
                    search_attribute=search_attribute
                    if search_attribute in {"sAMAccountName", "cn", "mail"}
                    else "sAMAccountName",
                    search_base=search_base,
                    auto_wildcard=True,
                    auto_select=False,
                    exclude=exclude,
                    title=title,
                )
            )
        except sdk_exceptions.ParameterError as e:
            if "has value *" in str(e):
                ui.output_error("Searching with only a wildcard is not supported for performance reasons, please try again.")
                continue
            raise e


@ldap_function
@idempotent_create
def add_to_group(domain_controller: DomainController, user_distinguishedName: str, group_distinguishedName: str) -> None:
    """Add a user to a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __user_distinguishedName:__ the distinguishedName of the user

    - __group_distinguishedName:__ the distinguishedName of the group

??? success "Output"
    Nothing is outputted by this action.

??? example
    Adding a user to a group.

    - The user & group are searched for using the _sAMAccountName_ attribute, & saved as new variables _user_ & _group_.

    - This action is then supplied the required _distinguishedName_ using fields from the _user_ & _group_ variables.
    ```YAML linenums="1"
    - active_directory.groups.search:
        sAMAccountName: example.group
      load:
        domain_controller: domain_controller_info
      save: group

    - active_directory.users.search:
        sAMAccountName: example.user
      load:
        domain_controller: domain_controller_info
      save: user

    - active_directory.users.add_to_group:
      load:
        user_distinguishedName: user.distinguishedName
        group_distinguishedName: group.distinguishedName
        domain_controller: domain_controller_info
    ```
"""
    groups.add_member(
        domain_controller, distinguished_name=group_distinguishedName, member_distinguished_name=user_distinguishedName
    )


@ldap_function
def add_to_groups_interactive(domain_controller: DomainController, distinguishedName: str) -> List[Group]:
    """Interactively add a user to one or more groups.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    A list of [Group Dictionaries](../inputs_outputs/#group) the user was added to.

??? example

    Triggering an interactive process where user _Example User_ can be added to one or more groups.

    ```YAML linenums="1"
    - active_directory.users.add_to_groups_interactive:
        distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return list(map(process_group, helpers.add_user_to_groups(domain_controller, distinguished_name=distinguishedName)[0]))


@ldap_function
@idempotent_remove
def remove_from_group(domain_controller: DomainController, user_distinguishedName: str, group_distinguishedName: str) -> None:
    """Remove a user from a group.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __user_distinguishedName:__ the distinguishedName of the user

    - __group_distinguishedName:__ The distinguishedName of the group

??? success "Output"
    Nothing is outputted by this action.

??? example
    Removing user _Example User_ from the group _Example Group_.
    ```YAML linenums="1"
    - active_directory.user.remove_from_group:
        user_distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
        group_distinguishedName: "CN=Example Group,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    groups.remove_member(
        domain_controller, distinguished_name=group_distinguishedName, member_distinguished_name=user_distinguishedName
    )


@ldap_function
def remove_from_groups_interactive(domain_controller: DomainController, distinguishedName: str) -> List[Group]:
    """Interactively remove a user from one or more groups.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    A list of [Group Dictionaries](../inputs_outputs/#group) the user was removed from.

??? example
    Triggering an interactive process where user _Example User_ can be removed from one or more groups.
    ```YAML linenums="1"
    - active_directory.users.remove_from_groups_interactive:
        distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return list(map(process_group, helpers.remove_user_from_groups(domain_controller, distinguished_name=distinguishedName)))


@ldap_function
@idempotent_create
def create(
    domain_controller: DomainController,
    distinguishedName: str,
    sAMAccountName: str,
    extra_params: Optional[Dict[str, Any]] = None,
) -> None:
    """Create a new user.

!!! question "Minimum Plugin Version: 1.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __sAMAccountName:__ the new user sAMAccountName

    - __distinguishedName:__ the new user distinguishedName

    - __extra_params:__ a dictionary containing any extra LDAP attributes & values for the user

??? success "Output"
    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - active_directory.users.create:
        sAMAccountName: example.user
        distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
        extra_params:
          info: An example user
          cn: Example User
      load:
        domain_controller: domain_controller_info
    ```
    """
    logger.info(f"creating user '{sAMAccountName}' at '{distinguishedName}' with the following extra parameters: {extra_params}")
    try:
        return users.create(
            domain_controller, distinguished_name=distinguishedName, samaccountname=sAMAccountName, **extra_params or {}
        )
    except sdk_exceptions.ParameterError as e:
        logger.error(e)
        message = "One or more of the following is not a supported user attribute: {}.".format(
            ", ".join([f"'{key}'" for key in list(extra_params.keys())])
        )
        ui.output_error(message)
        raise InvalidAttribute(message)


@ldap_function
@idempotent_remove
def delete(domain_controller: DomainController, distinguishedName: str) -> None:
    """Delete a user.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user to delete

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.delete:
        distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return users.delete(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def move(domain_controller: DomainController, distinguishedName: str, parent_distinguishedName: str) -> None:
    """Move a user to a different OU or CN.

!!! question "Minimum Plugin Version: 1.5.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user to move

    - __parent_distinguishedName:__ the distinguishedName of the OU or CN to move the user to

??? success "Output"

    Nothing is outputted by this action.

??? example

    Moving a user from the 'Users' CN to the 'Staff' OU:

    ```YAML linenums="1"
    - active_directory.users.move:
        distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
        parent_distinguishedName: "OU=Staff,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
    """
    return users.move(domain_controller, distinguished_name=distinguishedName, new_parent=parent_distinguishedName)


@ldap_function
def set_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: Union[str, int]) -> None:
    """Set a user attribute value.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to set

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.get_interactive:
      load:
        domain_controller: domain_controller_info
      save: user

    - active_directory.users.set_attribute:
        name: info
        value: Example Info Value
      load:
        distinguishedName: user.distinguishedName
        domain_controller: domain_controller_info
    ```
    """
    return users.set_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def set_attributes(domain_controller: DomainController, distinguishedName: str, attributes: AttributeSet) -> None:
    """Set multiple user attribute values.

This action clears & then sets each supplied attribute.

Multi-value string attributes are not supported by this action.

!!! question "Minimum Plugin Version: 7.11.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __attributes:__ a list of [AttributePairs](../inputs_outputs/#attributepair)

??? success "Output"

    Nothing is outputted by this action.

??? example

    ```YAML linenums="1"
    - active_directory.users.get_interactive:
      load:
        domain_controller: domain_controller_info
      save: user

    - active_directory.users.set_attributes:
        attributes:
          - name: info
            value: Example Info Value
          - name: description
            value: Example description
      load:
        distinguishedName: user.distinguishedName
        domain_controller: domain_controller_info
    ```
    """

    def _clear_and_set(name, value):
        clear_attribute(domain_controller, distinguishedName, name)
        set_attribute(domain_controller, distinguishedName, name, value)

    list(map(lambda attr_dict: _clear_and_set(attr_dict["name"], attr_dict["value"]), attributes))


@ldap_function
def replace_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: Union[str, int]) -> None:
    """Replace a user attribute value.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to set

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.get_interactive:
      load:
        domain_controller: domain_controller_info
      save: user

    - active_directory.users.replace_attribute:
        name: info
        value: Example Info Value
      load:
        distinguishedName: user.distinguishedName
        domain_controller: domain_controller_info
    ```
    """
    return users.replace_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def delete_attribute(domain_controller: DomainController, distinguishedName: str, name: str, value: Union[str, int]) -> None:
    """Delete a certain value from a user attribute.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __name:__ the name of the LDAP attribute

    - __value:__ the value to delete

??? success "Output"
    Nothing is outputted by this action.

??? example

    Clearing the existing _info_ value of a user.

    ```YAML linenums="1"
    - active_directory.users.get_interactive:
      load:
        domain_controller: domain_controller_info
      save: user

    - active_directory.users.delete_attribute:
        name: info
      load:
        distinguishedName: user.distinguishedName
        value: user.info
        domain_controller: domain_controller_info
    ```
    """
    return users.delete_attribute(domain_controller, distinguished_name=distinguishedName, name=name, value=value)


@ldap_function
def clear_attribute(domain_controller: DomainController, distinguishedName: str, name: str) -> None:
    """Clear a particular LDAP attribute on a user.

Works for single-valued & multi-valued string LDAP attributes.

!!! question "Minimum Plugin Version: 3.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __name:__ the name of the LDAP attribute

??? success "Output"

    Nothing is outputted by this action.

??? example

    Searching for a user & clearing their __carLicense__ field:

    ```YAML linenums="1"
    - active_directory.users.by_samaccountname:
        sAMAccountName: john.smith
      load:
        domain_controller: domain_controller
      save: user

    - active_directory.users.clear_attribute:
        name: carLicense
      load:
        distinguishedName: user.distinguishedName
        domain_controller: domain_controller
    ```
"""
    existing_value = by_distinguishedname(domain_controller, distinguishedName)[name]

    if existing_value is None:
        # Nothing to do as the attribute has no value.
        return

    delete_func = functools.partial(delete_attribute, domain_controller, distinguishedName, name=name)
    list(map(lambda v: delete_func(value=v), existing_value)) if isinstance(existing_value, list) else delete_func(
        value=existing_value
    )


@ldap_function
def set_password(domain_controller: DomainController, distinguishedName: str, password: str) -> None:
    """Set a user's password.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __password:__ the new password

??? success "Output"

    Nothing is outputted by this action.

??? example
    - Setting the password for user _John Smith_ in the root _Users_ CN

    - The PPA UI _input_password_ action is used to get & save the new password

    ```YAML linenums="1"
    - ppa.ui.input_password:
        text: New Password
      save: new_password

    - active_directory.users.set_password:
        distinguishedName: "CN=John Smith,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
        password: new_password
    ```
"""
    return users.set_password(domain_controller, distinguished_name=distinguishedName, new_password=password)


def set_password_interactive(domain_controller: DomainController, distinguishedName: str) -> str:
    """Interactively set a user's password. This operation:

- Asks the task operator to supply & confirm a new password.
- Attempts to set the password for the user.
- If the password is refused by Active Directory the operation will be repeated repeated.

!!! question "Minimum Plugin Version: 5.6.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary
    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"

    The password that was successfully set for the user.

??? example
    ```YAML linenums="1"
    - active_directory.users.set_password_interactive:
        distinguishedName: "CN=John Smith,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
"""

    def _get_password():
        while True:
            password = ui.input_password("New Password", meter=True, element_id="new_password")
            if password == ui.input_password("Confirm Password", meter=True, element_id="confirm_password"):
                break
            else:
                ui.output_error("Passwords must match.", element_id="password_error")
        return password

    def _set_password(password):
        try:
            set_password(domain_controller, distinguishedName, password)
            return True
        except exceptions.InvalidPassword as e:
            ui.output_error(str(e), element_id="invalid_password_exception")
            ui.output_error("Please try again.", element_id="invalid_password_error")
            return False

    for _ in range(0, 10):
        password = _get_password()
        if _set_password(password):
            return password
    raise Exception("Failed to set password after 10 attempts. Operation failed.")


@ldap_function
def disable(domain_controller: DomainController, distinguishedName: str) -> None:
    """Disable a user.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.disable:
        distinguishedName: "CN=John Smith,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
"""
    return users.disable(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def enable(domain_controller: DomainController, distinguishedName: str) -> None:
    """Enable a user.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.enable:
        distinguishedName: "CN=John Smith,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
"""
    return users.enable(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def unlock(domain_controller: DomainController, distinguishedName: str) -> None:
    """Unlock a user.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.unlock:
        distinguishedName: "CN=John Smith,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
"""
    return users.unlock(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def force_password_change(domain_controller: DomainController, distinguishedName: str) -> None:
    """Set the 'force password change at next logon' flag against a user.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"
    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.force_password_change:
        distinguishedName: "CN=John Smith,CN=Users,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
    ```
"""
    return users.force_password_change(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def get_all(
    domain_controller: DomainController, search_base: Optional[str] = None, one_level: Optional[bool] = False
) -> List[User]:
    """Get all users.

!!! danger "This operation can be slow on large domains"

!!! question "Minimum Plugin Version: 6.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

    - __one_level:__ set to `true` to only search one level (see [here](../advanced_queries/#search-scopes) for more information)

??? success "Output"

    A list of [Users](../inputs_outputs/#user).

??? example

    ```YAML linenums="1"
    - active_directory.users.get_all:
      load:
        domain_controller: domain_controller_info
      save: all_users
    ```
"""
    return list(
        map(
            process_user, users.get_all(domain_controller, search_base=search_base, search_scope=bool_to_search_scope(one_level)),
        )
    )


@ldap_function
def get_enabled(
    domain_controller: DomainController, search_base: Optional[str] = None, one_level: Optional[bool] = False
) -> List[User]:
    """Get enabled users.

!!! danger "This operation can be slow on large domains"

!!! question "Minimum Plugin Version: 7.12.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

    - __one_level:__ set to `true` to only search one level (see [here](../advanced_queries/#search-scopes) for more information)

??? success "Output"

    A list of enabled [Users](../inputs_outputs/#user).

??? example

    ```YAML linenums="1"
    - active_directory.users.get_enabled:
      load:
        domain_controller: domain_controller_info
      save: enabled_users
    ```
"""
    return list(filter(lambda u: u["is_enabled"], get_all(domain_controller, search_base=search_base, one_level=one_level)))


@ldap_function
def get_disabled(
    domain_controller: DomainController, search_base: Optional[str] = None, one_level: Optional[bool] = False
) -> List[User]:
    """Get disabled users.

!!! danger "This operation can be slow on large domains"

!!! question "Minimum Plugin Version: 7.12.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __search_base:__ optional start point for the search (see [here](../advanced_queries/#search-location) for more information)

    - __one_level:__ set to `true` to only search one level (see [here](../advanced_queries/#search-scopes) for more information)

??? success "Output"

    A list of disabled [Users](../inputs_outputs/#user).

??? example

    ```YAML linenums="1"
    - active_directory.users.get_disabled:
      load:
        domain_controller: domain_controller_info
      save: disabled_users
    ```
"""
    return list(filter(lambda u: u["is_disabled"], get_all(domain_controller, search_base=search_base, one_level=one_level)))


@ldap_function
def group_memberships(domain_controller: DomainController, distinguishedName: str, nested: Optional[bool] = False) -> List[Group]:
    """Get the group memberships of a user.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __nested:__ Set to `true` to include nested group memberships in the search

??? success "Output"
    A list of group dictionaries.

??? warning
    Getting nested group memberships can be slow.

??? example
    - Getting all group memberships for user _Example User_

    - Saving the results as a new variable called `group_memberships`

    ```YAML linenums="1"
    - active_directory.users.group_memberships:
        distinguishedName: "CN=Example User,CN=Users,DC=Example,DC=Domain"
        nested: true
      load:
        domain_controller: domain_controller_info
      save: group_memberships
    ```
    """
    return list(
        map(process_group, users.group_memberships(domain_controller, distinguished_name=distinguishedName, nested=nested))
    )


def test_credentials(address: str, port: str, domain: str, username: str, password: str) -> None:
    """Test a user's credentials. Often used after resetting a password.

Takes the same inputs a [DomainController](../inputs_outputs/#domaincontroller) dictionary contains, but each input is supplied separately.

!!! question "Minimum Plugin Version: 1.0.0"

??? info "Input"

    - __address__: the Domain Controller IP or DNS address

    - __domain__: the FQDN of the Active Directory domain

    - __username__: username for authentication

    - __password__: password for authentication

??? success "Output"
    Nothing is outputted by this action.

??? example
    Testing a new password for user _John Smith_.

    - The `new_password` variable was set in an earlier step (not shown)

    - Note the `domain_controller` dictionary is used to supply the `address`, `port`, & `domain` keys

    ```YAML linenums="1"
    - active_directory.users.test_credentials:
        username: john.smith
      load:
        address: domain_controller.address
        port: domain_controller.port
        domain: domain_controller.domain
        password: new_password
    ```
"""
    return users.test_credentials(address=address, port=port, domain=domain, username=username, password=password)


def output_table(text: str, users: Union[User, List[User]]) -> None:
    """Display users in a table.

The table will have the following columns:

- Common Name
- SAM Account Name
- Email Address
- Enabled
- Locked

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display](#display).

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ a single or list of [User Dictionaries](../inputs_outputs/#user)

??? success "Output"
    Nothing is outputted by this action.

??? example
    - Getting all users whose common names start with _Test_

    - Saving the results as a new variable called `test_users`

    - Using this action to show the `test_users` in a table

    ```YAML linenums="1"
    - active_directory.users.search:
        cn: Test*
      load:
        domain_controller: domain_controller_info
      save: test_users

    - active_directory.users.output_table:
        text: Test User Accounts
      load:
        users: test_users
    ```
    """
    users = [users] if not isinstance(users, list) else users
    ui.output_table(
        text,
        header=USER_TABLE_HEADER,
        data=[[str(user.get(field)) if user.get(field) is not None else "" for field in USER_TABLE_FIELDS] for user in users],
    )


def display(
    text: str, users: Union[User, List[User]], header: Optional[List[str]] = None, fields: Optional[List[str]] = None,
) -> None:
    """Display users in a table.

The table will have the following columns by default:

- Common Name
- SAM Account Name
- Email Address
- Enabled
- Locked

These can be customised by supplying the **header** & **fields** inputs.

!!! question "Minimum Plugin Version: 7.12.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

    - __header:__ optional table header (see default above)

    - __fields:__ optional list of [User](../inputs_outputs/#user) keys (see default above)

??? success "Output"

    Nothing is outputted by this action.

??? example
    ```YAML linenums="1"
    - active_directory.users.search:
        cn: Test*
      load:
        domain_controller: domain_controller_info
      save: test_users

    - active_directory.users.display:
        text: Test User Accounts
      load:
        users: test_users
    ```
    """
    header, fields = _process_header_and_fields(header, fields)
    tables.output_table(text, header=header or USER_TABLE_HEADER, extract=fields or USER_TABLE_FIELDS, records=users)


def display_tabbed(text: str, tabs: Dict[str, List[User]]) -> None:
    """Display users in a table with multiple tabs.

This action can be used to display user lists from multiple Active Directories.

The table will have the following columns:

- Common Name
- SAM Account Name
- Email Address
- Enabled
- Locked

!!! question "Minimum Plugin Version: 6.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is any number of [Users](../inputs_outputs/#user)

??? success "Output"

    Nothing is outputted by this action.

??? example

    Auditing user accounts from 2 domains & presenting them in a tabbed table:

    ```YAML linenums="1"
    - set:
        name: tabs
        value: {}

    - active_directory.users.get_all:
      load:
        domain_controller: production_domain
      save: production_users

    - active_directory.users.get_all:
      load:
        domain_controller: uat_domain
      save: uat_users

    - ppa_tools.dictionaries.insert:
        name: Production Active Directory
      load:
        value: production_users
        dictionary: tabs
      save: tabs

    - ppa_tools.dictionaries.insert:
        name: UAT Active Directory
      load:
        value: uat_users
        dictionary: tabs
      save: tabs

    - active_directory.users.display_tabbed:
        text: Active Directory Users
      load:
        tabs: tabs
    ```
    """
    tables.output_tabbed_table(text, header=USER_TABLE_HEADER, extract=USER_TABLE_FIELDS, tabs=tabs)


def output_tabbed_table(text: str, tabs: Dict[str, List[User]]) -> None:
    """Display users in a table with multiple tabs.

This action can be used to display user lists from multiple Active Directories.

The table will have the following columns:

- Common Name
- SAM Account Name
- Email Address
- Enabled
- Locked

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [display_tabbed](#display_tabbed).

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __tabs:__ a dictionary where each key is a tab name & each value is a list of [Users](../inputs_outputs/#user)

??? success "Output"

    Nothing is outputted by this action.

??? example

    Auditing user accounts from 2 domains & presenting them in a tabbed table:

    ```YAML linenums="1"
    - set:
        name: tabbed_table
        value: {}

    - active_directory.users.get_all:
      load:
        domain_controller: production_domain
      save: production_users

    - active_directory.users.get_all:
      load:
        domain_controller: uat_domain
      save: uat_users

    - ppa_tools.dictionaries.insert:
        name: Production Active Directory
      load:
        value: production_users
        dictionary: tabbed_table
      save: tabbed_table

    - ppa_tools.dictionaries.insert:
        name: UAT Active Directory
      load:
        value: uat_users
        dictionary: tabbed_table
      save: tabbed_table

    - active_directory.users.output_tabbed_table:
        text: Active Directory Users
      load:
        tabs: tabbed_table
    ```
    """
    ui.output_table(
        text,
        header=USER_TABLE_HEADER,
        data={
            key: [[str(user.get(field)) if user.get(field) is not None else "" for field in USER_TABLE_FIELDS] for user in users]
            for key, users in tabs.items()
        },
    )


def output_custom_table(text: str, header: List[str], attributes: List[str], users: Union[User, List[User]]) -> None:
    """Display users in a table using custom headers & LDAP attributes.

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __header:__ a list of column names

    - __attributes:__ a list of LDAP attributes (one for each column name in the `header`)

    - __users:__ any number of [Users](../inputs_outputs/#user)

??? success "Output"
    Nothing is outputted by this action.

??? example
    Showing `mail`, `distinguishedName`, & `userAccountControl` for users in the variable `test_users`.

    ```YAML linenums="1"
    - active_directory.users.output_custom_table:
        text: Custom Users Table
        header:
          - Email Address
          - Distinguished Name
          - User Account Control
        attributes:
          - mail
          - distinguishedName
          - userAccountControl
      load:
        users: test_users
    ```
    """
    users = [users] if not isinstance(users, list) else users
    ui.output_table(
        text,
        header=header,
        data=[[str(user.get(attr)) if user.get(attr) is not None else "" for attr in attributes] for user in users],
    )


def input_table(
    text: str, users: Union[User, List[User]], minimum: Optional[int] = None, maximum: Optional[int] = None
) -> List[User]:
    """Display users in a table, & allow the task operator to make a selection.

The table will have the following columns:

- Common Name
- SAM Account Name
- Email Address
- Enabled
- Locked

!!! warning "Deprecation Warning"

    This action will soon be deprecated in favour of [select](#select).

!!! question "Minimum Plugin Version: 5.0.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ a single or list of [User Dictionaries](../inputs_outputs/#user)

    - __minimum:__ the minimum number of acceptable selections

    - __maximum:__ the maximum number of acceptable selections

??? success "Output"

    A list of [User Dictionaries](../inputs_outputs/#user).

??? tip

    - If neither a _minimum_ or _maximum_ is provided, the task operator will be able to submit 0 selections.
    - If _minimum_ or _maximum_ are provided, the operation will repeat until the task operator makes a valid number
        of selections.

??? example

    - Getting all users whose common names start with _Test_

    - Saving the results as a new variable called `test_users`

    - Using this action to show the `test_users` in a table, requiring the task operator selects at least 1

    - The selection is saved as a new variable called `selected_users`

    ```YAML linenums="1"
    - active_directory.users.search:
        cn: Test*
      load:
        domain_controller: domain_controller_info
      save: test_users

    - active_directory.users.input_table:
        text: "Please Select >= 1 User(s)"
        minimum: 1
      load:
        users: test_users
      save: selected_users
    ```
    """
    users = [users] if not isinstance(users, list) else users
    table_element_id = str(uuid.uuid4())
    error_element_id = str(uuid.uuid4())
    while True:
        indices = ui.input_table(
            text,
            header=USER_TABLE_HEADER,
            data=[[str(user.get(field)) if user.get(field) is not None else "" for field in USER_TABLE_FIELDS] for user in users],
            element_id=table_element_id,
        )
        count = len(indices)
        if minimum and count < minimum:
            ui.output_error(f"Please choose at least {minimum} record(s)", element_id=error_element_id)
            continue
        if maximum and count > maximum:
            ui.output_error(f"Please choose no more than {maximum} record(s)", element_id=error_element_id)
            continue
        break

    return [users[i] for i in indices]


def select(
    text: str,
    users: Union[User, List[User]],
    minimum: Optional[int] = None,
    maximum: Optional[int] = None,
    header: Optional[List[str]] = None,
    fields: Optional[List[str]] = None,
) -> tables.TableSelection:
    """Display users in a table & prompt the task operator to make a selection.

The table will have the following columns by default:

- Common Name
- SAM Account Name
- Email Address
- Enabled
- Locked

These can be customised by supplying the **header** & **fields** inputs.

!!! question "Minimum Plugin Version: 7.12.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

    - __minimum:__ the minimum number of selections

    - __maximum:__ the maximum number of selections

    - __header:__ optional table header (see default above)

    - __fields:__ optional list of [User](../inputs_outputs/#user) keys (see default above)

??? success "Output"

    A [Selection](../../../tables/#selection) containing:

    - **total**: the number of selected [Users](../inputs_outputs/#user)

    - **all**: a list of selected [Users](../inputs_outputs/#user)

    - **first**: the first selected [User](../inputs_outputs/#user)

??? example

    ```YAML linenums="1"
    - active_directory.users.get_all:
      load:
        domain_controller: domain_controller
      save: users

    - active_directory.users.select:
        text: Select Users
      load:
        users: users
      save: selection
    ```
    """

    header, fields = _process_header_and_fields(header, fields)
    return tables.input_table(
        text,
        header=header or USER_TABLE_HEADER,
        extract=fields or USER_TABLE_FIELDS,
        records=users,
        minimum=minimum,
        maximum=maximum,
    )


def select_one(
    text: str, users: Union[User, List[User]], header: Optional[List[str]] = None, fields: Optional[List[str]] = None,
) -> User:
    """Display users in a table & prompt the task operator to select one.

The table will have the following columns:

- Common Name
- SAM Account Name
- Email Address
- Enabled
- Locked

These can be customised by supplying the **header** & **fields** inputs.

!!! question "Minimum Plugin Version: 7.12.0"

??? info "Input"

    - __text:__ the title of the table

    - __users:__ any number of [Users](../inputs_outputs/#user)

    - __header:__ optional table header (see default above)

    - __fields:__ optional list of [User](../inputs_outputs/#user) keys (see default above)

??? success "Output"

    A single [User](../inputs_outputs/#user).

??? example

    ```YAML linenums="1"
    - active_directory.users.get_all:
      load:
        domain_controller: domain_controller
      save: users

    - active_directory.users.select_one:
        text: Select a User
      load:
        users: users
      save: user
    ```
    """
    header, fields = _process_header_and_fields(header, fields)
    return select(text, users, minimum=1, maximum=1, header=header, fields=fields)["first"]


@ldap_function
def exists(domain_controller: DomainController, **search_params: Union[str, int, None]) -> bool:
    """Search for users using LDAP attributes & values to identify if any were found.

!!! question "Minimum Plugin Version: 1.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __search_params:__ a dictionary containing [user keys](../inputs_outputs#user) & values to use in the search


??? success "Output"

    A boolean is outputted by this action

    - `true` if one or more users are found

    - `false` if no users are found


??? example

    Searching using a unique attribute:

    ```YAML linenums="1"
    - active_directory.users.exists:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: item_exists
    ```

    Searching using a non-unique attribute:
    ```YAML linenums="1"
    - active_directory.users.exists:
        cn: "Maddison*"
      load:
        domain_controller: domain_controller_info
      save: item_exists
    ```

!!! tip "Wildcard Searching"

    It is possible to use `*` as a wildcard at the __end__ of search values, but this can make the search slow.

    """
    return bool(users.search(domain_controller, **search_params))


@ldap_function
def has_reversible_password(domain_controller: DomainController, distinguishedName: str) -> bool:
    """Determine if the supplied user's password is stored using reversible encryption.

!!! question "Minimum Plugin Version: 4.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user


??? success "Output"

    A boolean is outputted by this action

    - `true` if the password is stored using reversible encryption

    - `false` if the password is _not_ stored using reversible encryption


??? example

    ```YAML linenums="1"
    - active_directory.users.has_reversible_password:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: reversible
    ```
    """
    return users.has_reversible_password(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def smart_card_required(domain_controller: DomainController, distinguishedName: str) -> bool:
    """Determine if the supplied user account requires a smart card for interactive login.

!!! question "Minimum Plugin Version: 4.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user


??? success "Output"

    A boolean is outputted by this action

    - `true` if a smart card is required

    - `false` if a smart card is _not_ required


??? example

    ```YAML linenums="1"
    - active_directory.users.smart_card_required:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: card_required
    ```
    """
    return users.smart_card_required(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def password_will_expire(domain_controller: DomainController, distinguishedName: str) -> bool:
    """Determine if the supplied user account's password has an expiry.

!!! question "Minimum Plugin Version: 4.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user


??? success "Output"

    A boolean is outputted by this action

    - `true` if the user's password will expire

    - `false` if the user's password will _not_ expire


??? example

    ```YAML linenums="1"
    - active_directory.users.password_will_expire:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: password_will_expire
    ```
    """
    return users.password_will_expire(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def password_has_expired(
    domain_controller: DomainController, distinguishedName: str, maximum_password_age: Optional[Dict[str, int]] = None
) -> bool:
    """Determine if the supplied user account's password has expired.

!!! question "Minimum Plugin Version: 7.1.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __maximum_password_age:__ optional maximum password age outputted by [this action](../active_directory.domain/#maximum_password_age)

!!! tip "Running Against Multiple Users"

    If the **maximum_password_age** input is not supplied, the action will get it from the domain.

    If your task runs this action on many users, we recommended supplying this input.

    While this has only a small impact on performance, it will result in far fewer connections to Active Directory.

??? success "Output"

    A boolean is outputted by this action

    - `true` if the user's password has expired

    - `false` if the user's password has _not_ expired

??? example

    ```YAML linenums="1"
    - active_directory.users.password_has_expired:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: password_expired
    ```
    """
    try:
        return (
            get_password_expiry(domain_controller, distinguishedName, maximum_password_age=maximum_password_age)["total_seconds"]
            <= 0
        )
    except TypeError:  # The user's password will not expire.
        return False


@ldap_function
def password_not_required(domain_controller: DomainController, distinguishedName: str) -> bool:
    """Determine if the supplied user account can have a blank password.

!!! question "Minimum Plugin Version: 4.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user


??? success "Output"

    A boolean is outputted by this action

    - `true` if the user's password can be blank

    - `false` if the user's password _cannot_ be blank


??? example

    ```YAML linenums="1"
    - active_directory.users.password_not_required:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: password_not_required
    ```
    """
    return users.password_not_required(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def can_be_delegated(domain_controller: DomainController, distinguishedName: str) -> bool:
    """Determine if the supplied user account __has not__ been flagged as 'sensitive and not for delegation'.

!!! question "Minimum Plugin Version: 4.0.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user


??? success "Output"

    A boolean is outputted by this action

    - `true` if the user can be delegated

    - `false` if the user cannot be delegated


??? example

    ```YAML linenums="1"
    - active_directory.users.can_be_delegated:
        distinguishedName: "CN=Full Name,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: can_be_delegated
    ```
    """
    return users.can_be_delegated(domain_controller, distinguished_name=distinguishedName)


@ldap_function
def get_free_uid_numbers(
    domain_controller: DomainController, start: int, end: int, quantity: int, block_size: int = 5
) -> List[int]:
    """Get **uidNumber**s in a supplied range that _are not_ assigned to any users.

!!! question "Minimum Plugin Version: 4.2.0"


??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __start:__ the start of the range

    - __end:__ the end of the range

    - __quantity:__ the number of free uidNumbers you require

    - __block_size:__ how many numbers to query at once (defaults to `5`)


??? tip "Search Performance"

    This action queries the domain controller for each block of numbers in the supplied range.

    If this action performs slowly, use a higher `block_size` to reduce the number of queries.

    The action stops as soon as the supplied `quantity` is met, so use the lowest acceptable `quantity` for best results.


??? success "Output"

    A list of numbers in the supplied range that are _not_ used as any user's **uidNumber**.


??? example

    Getting 100 unused uidNumbers between 20,000 & 30,000:

    ```YAML linenums="1"
    - active_directory.users.get_free_uid_numbers:
        start: 20000
        end: 30000
        quantity: 100
      load:
        domain_controller: domain_controller_info
      save: free_uid_numbers
    ```
    """

    if quantity > (end - start):
        message = f"Supplied quantity {quantity} cannot be fulfilled by range {start}-{end}."
        ui.output_error(message)
        raise ValueError(message)

    free_uids = []

    def _get_chunks(sequence):
        return (sequence[index : (index + block_size)] for index in range(0, len(sequence), block_size))

    for uid_range in _get_chunks(range(start, end + 1)):
        needed = quantity - len(free_uids)
        uid_range = list(uid_range)
        search_filter = f"(&(uidNumber>={uid_range[0]})(uidNumber<={uid_range[-1]}))"
        search_query = f"(&{_users.USER_BASE_QUERY}{search_filter})"

        if users_with_uids := _ldaps._search(domain_controller, search_query):
            if len(users_with_uids) == len(uid_range):
                # All full, move on...
                continue
            used = list(map(lambda u: int(u["uidNumber"][0].decode("utf-8")), users_with_uids))
            free = list(filter(lambda uid: uid not in used, uid_range))
        else:
            # All the UIDs in the current chunk are free.
            free = uid_range

        if len(free) > needed:
            # Only grab the quantity we need.
            free = free[:needed]

        logger.debug(f"adding {len(free)} free uids from current chunk of {len(uid_range)}")
        free_uids.extend(free)

        if len(free_uids) == quantity:
            return free_uids

    else:
        message = (
            f"Only {len(free_uids)} free uidNumbers were found between {start}-{end}, & the required quantity is {quantity}."
        )
        ui.output_error(message)
        raise IndexError(message)


def get_password_expiry(
    domain_controller: DomainController, distinguishedName: str, maximum_password_age: Optional[Dict[str, int]] = None
) -> Optional[PasswordExpiry]:
    """Find out when a user's password will expire.

!!! question "Minimum Plugin Version: 7.0.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

    - __maximum_password_age:__ optional maximum password age outputted by [this action](../active_directory.domain/#maximum_password_age)

!!! tip "Running Against Multiple Users"

    If the **maximum_password_age** input is not supplied, the action will get it from the domain.

    If your task runs this action on many users, we recommended supplying this input.

    While this has only a small impact on performance, it will result in far fewer connections to Active Directory.

??? success "Output"

    A single [PasswordExpiry](../inputs_outputs/#passwordexpiry) or **null** if either of the following are true:

    - The domain has no maximum password age

    - The user does not have a password set

??? example "Example 1 - Getting Single Expiry"

    {% raw %}
    ```YAML linenums="1"
    - active_directory.users.get_password_expiry:
        distinguishedName: CN=Full Name,OU=Example,DC=Example,DC=Domain
      load:
        domain_controller: domain_controller_info
      save: expiry

    - ppa.ui.output_markdown:
        doc: >
          The user's password will expire on {{ expiry.timestamp }}.

          Time remaining:

          **Days:** {{ expiry.days }}

          **Hours:** {{ expiry.hours }}

          **Minutes:** {{ expiry.minutes }}
    ```
    {% endraw %}


??? example "Example 2 - Getting Multiple Expiries"

    - Getting the maximum password age separately & supplying it to this action

    - Running the action against a **sequence** of users

    - Supplying the maximum password age to the action means it is only fetched once

    ```YAML linenums="1" hl_lines="4 9 11"
    - active_directory.domain.get_maximum_password_age:
      load:
        domain_controller: domain_controller
      save: maximum_password_age

    - active_directory.users.get_password_expiry:
      load:
        distinguishedName: loop.action.item.value.distinguishedName
        maximum_password_age: maximum_password_age
        domain_controller: domain_controller_info
      sequence: users
      save: user_password_expiries
    ```
    """

    if maximum_password_age is None:
        maximum_password_age = domain.maximum_password_age(domain_controller)
        if maximum_password_age is None:
            logger.debug("password will never expire as the domain has no maximum password age")
            return None

    user = by_distinguishedname(domain_controller, distinguishedName)
    if user["password_age"] is None:
        logger.debug("password will never expire as the user pwdLastSet field has no value")
        return None

    # Both the maximum password age & current password age can be unpacked to create time deltas. All we need
    # to do is subtract the smaller delta (current password age) from the larger delta (maximum password age)
    # to get another time delta expressing the difference between them. That will be the remaining time until
    # the password expires, & it can be conveniently expressed as a timestamp by adding it to "now".
    remaining_time_delta = datetime.timedelta(**maximum_password_age) - datetime.timedelta(**user["password_age"])
    remaining_seconds = int(remaining_time_delta.total_seconds())
    return PasswordExpiry(
        timestamp=f"{(datetime.datetime.utcnow() + remaining_time_delta).strftime('%H:%M:%S %d/%m/%Y')} UTC",
        total_seconds=remaining_seconds,
        **_translators.seconds_to_units(remaining_seconds),
    )


def get_uac_properties(domain_controller: DomainController, distinguishedName: str) -> List[str]:
    """Get a list of the supplied user's UAC properties.

!!! question "Minimum Plugin Version: 7.5.0"

??? info "Input"

    - __domain_controller:__ a [DomainController](../inputs_outputs/#domaincontroller) dictionary

    - __distinguishedName:__ the distinguishedName of the user

??? success "Output"

    A list containing one or more of the following UAC properties assigned to the user.

    See more about UAC property flags [here](https://docs.microsoft.com/en-us/troubleshoot/windows-server/identity/useraccountcontrol-manipulate-account-properties).

    - __SCRIPT__

    - __ACCOUNTDISABLE__

    - __HOMEDIR_REQUIRED__

    - __LOCKOUT__

    - __PASSWD_NOTREQD__

    - __PASSWD_CANT_CHANGE__

    - __ENCRYPTED_TEXT_PWD_ALLOWED__

    - __TEMP_DUPLICATE_ACCOUNT__

    - __NORMAL_ACCOUNT__

    - __INTERDOMAIN_TRUST_ACCOUNT__

    - __WORKSTATION_TRUST_ACCOUNT__

    - __SERVER_TRUST_ACCOUNT__

    - __DONT_EXPIRE_PASSWORD__

    - __MNS_LOGON_ACCOUNT__

    - __SMARTCARD_REQUIRED__

    - __TRUSTED_FOR_DELEGATION__

    - __NOT_DELEGATED__

    - __USE_DES_KEY_ONLY__

    - __DONT_REQ_PREAUTH__

    - __PASSWORD_EXPIRED__

    - __TRUSTED_TO_AUTH_FOR_DELEGATION__

    - __PARTIAL_SECRETS_ACCOUNT__

??? example
    {% raw %}
    ```YAML linenums="1"
    - active_directory.users.get_uac_properties:
        distinguishedName: "CN=Example User,OU=Example,DC=Example,DC=Domain"
      load:
        domain_controller: domain_controller_info
      save: uac_properties

    - ppa.ui.output_markdown:
        doc: >
          ### UAC Properties

          **User:** Example User

          **Flags:**

          {% for property in uac_properties %}

          - {{ property }}

          {% endfor %}
    ```
    {% endraw %}
    """
    uac_bitmask = by_distinguishedname(domain_controller, distinguishedName)["userAccountControl"]
    return [name for name, value in UACPropertyFlags.items() if value | uac_bitmask == uac_bitmask]
